@@ -1,9 +1,12 @@
 'use strict';
 
+/* global WebAssembly */
+
 const {
+  JSON,
+  Object,
   SafeMap,
-  StringPrototype,
-  JSON
+  StringPrototype
 } = primordials;
 
 const { NativeModule } = require('internal/bootstrap/loaders');
@@ -72,11 +75,11 @@ translators.set('commonjs', async function commonjsStrategy(url, isMain) {
   ];
   if (module && module.loaded) {
     const exports = module.exports;
-    return createDynamicModule(['default'], url, (reflect) => {
+    return createDynamicModule([], ['default'], url, (reflect) => {
       reflect.exports.default.set(exports);
     });
   }
-  return createDynamicModule(['default'], url, () => {
+  return createDynamicModule([], ['default'], url, () => {
     debug(`Loading CJSModule ${url}`);
     // We don't care about the return val of _load here because Module#load
     // will handle it for us by checking the loader registry and filling the
@@ -97,7 +100,7 @@ translators.set('builtin', async function builtinStrategy(url) {
   }
   module.compileForPublicLoader(true);
   return createDynamicModule(
-    [...module.exportKeys, 'default'], url, (reflect) => {
+    [], [...module.exportKeys, 'default'], url, (reflect) => {
       debug(`Loading BuiltinModule ${url}`);
       module.reflect = reflect;
       for (const key of module.exportKeys)
@@ -116,11 +119,21 @@ translators.set('json', async function jsonStrategy(url) {
   let module = CJSModule._cache[modulePath];
   if (module && module.loaded) {
     const exports = module.exports;
-    return createDynamicModule(['default'], url, (reflect) => {
+    return createDynamicModule([], ['default'], url, (reflect) => {
       reflect.exports.default.set(exports);
     });
   }
   const content = await readFileAsync(pathname, 'utf-8');
+  // A require call could have been called on the same file during loading and
+  // that resolves synchronously. To make sure we always return the identical
+  // export, we have to check again if the module already exists or not.
+  module = CJSModule._cache[modulePath];
+  if (module && module.loaded) {
+    const exports = module.exports;
+    return createDynamicModule(['default'], url, (reflect) => {
+      reflect.exports.default.set(exports);
+    });
+  }
   try {
     const exports = JsonParse(stripBOM(content));
     module = {
@@ -136,8 +149,32 @@ translators.set('json', async function jsonStrategy(url) {
     throw err;
   }
   CJSModule._cache[modulePath] = module;
-  return createDynamicModule(['default'], url, (reflect) => {
+  return createDynamicModule([], ['default'], url, (reflect) => {
     debug(`Parsing JSONModule ${url}`);
     reflect.exports.default.set(module.exports);
   });
 });
+
+// Strategy for loading a wasm module
+translators.set('wasm', async function(url) {
+  const pathname = fileURLToPath(url);
+  const buffer = await readFileAsync(pathname);
+  debug(`Translating WASMModule ${url}`);
+  let compiled;
+  try {
+    compiled = await WebAssembly.compile(buffer);
+  } catch (err) {
+    err.message = pathname + ': ' + err.message;
+    throw err;
+  }
+
+  const imports =
+      WebAssembly.Module.imports(compiled).map(({ module }) => module);
+  const exports = WebAssembly.Module.exports(compiled).map(({ name }) => name);
+
+  return createDynamicModule(imports, exports, url, (reflect) => {
+    const { exports } = new WebAssembly.Instance(compiled, reflect.imports);
+    for (const expt of Object.keys(exports))
+      reflect.exports[expt].set(exports[expt]);
+  });
+});
