@@ -1,5 +1,5 @@
-#include "node_internals.h"
-#include "node_watchdog.h"
+#include "node_errors.h"
+#include "util-inl.h"
 #include "base_object-inl.h"
 
 namespace node {
@@ -7,16 +7,18 @@ namespace util {
 
 using v8::ALL_PROPERTIES;
 using v8::Array;
+using v8::ArrayBufferView;
 using v8::Boolean;
 using v8::Context;
+using v8::Function;
 using v8::FunctionCallbackInfo;
 using v8::FunctionTemplate;
+using v8::Global;
 using v8::IndexFilter;
 using v8::Integer;
 using v8::Isolate;
 using v8::KeyCollectionMode;
 using v8::Local;
-using v8::NewStringType;
 using v8::Object;
 using v8::ONLY_CONFIGURABLE;
 using v8::ONLY_ENUMERABLE;
@@ -25,7 +27,6 @@ using v8::Private;
 using v8::Promise;
 using v8::PropertyFilter;
 using v8::Proxy;
-using v8::ReadOnly;
 using v8::SKIP_STRINGS;
 using v8::SKIP_SYMBOLS;
 using v8::String;
@@ -57,6 +58,16 @@ static void GetOwnNonIndexProperties(
   args.GetReturnValue().Set(properties);
 }
 
+static void GetConstructorName(
+    const FunctionCallbackInfo<Value>& args) {
+  CHECK(args[0]->IsObject());
+
+  Local<Object> object = args[0].As<Object>();
+  Local<String> name = object->GetConstructorName();
+
+  args.GetReturnValue().Set(name);
+}
+
 static void GetPromiseDetails(const FunctionCallbackInfo<Value>& args) {
   // Return undefined if it's not a Promise.
   if (!args[0]->IsPromise())
@@ -65,13 +76,13 @@ static void GetPromiseDetails(const FunctionCallbackInfo<Value>& args) {
   auto isolate = args.GetIsolate();
 
   Local<Promise> promise = args[0].As<Promise>();
-  Local<Array> ret = Array::New(isolate, 2);
 
   int state = promise->State();
-  ret->Set(0, Integer::New(isolate, state));
+  Local<Value> values[2] = { Integer::New(isolate, state) };
+  size_t number_of_values = 1;
   if (state != Promise::PromiseState::kPending)
-    ret->Set(1, promise->Result());
-
+    values[number_of_values++] = promise->Result();
+  Local<Array> ret = Array::New(isolate, values, number_of_values);
   args.GetReturnValue().Set(ret);
 }
 
@@ -82,11 +93,13 @@ static void GetProxyDetails(const FunctionCallbackInfo<Value>& args) {
 
   Local<Proxy> proxy = args[0].As<Proxy>();
 
-  Local<Array> ret = Array::New(args.GetIsolate(), 2);
-  ret->Set(0, proxy->GetTarget());
-  ret->Set(1, proxy->GetHandler());
+  Local<Value> ret[] = {
+    proxy->GetTarget(),
+    proxy->GetHandler()
+  };
 
-  args.GetReturnValue().Set(ret);
+  args.GetReturnValue().Set(
+      Array::New(args.GetIsolate(), ret, arraysize(ret)));
 }
 
 static void PreviewEntries(const FunctionCallbackInfo<Value>& args) {
@@ -98,20 +111,16 @@ static void PreviewEntries(const FunctionCallbackInfo<Value>& args) {
   Local<Array> entries;
   if (!args[0].As<Object>()->PreviewEntries(&is_key_value).ToLocal(&entries))
     return;
-  // Fast path for WeakMap, WeakSet and Set iterators.
+  // Fast path for WeakMap and WeakSet.
   if (args.Length() == 1)
     return args.GetReturnValue().Set(entries);
-  Local<Array> ret = Array::New(env->isolate(), 2);
-  ret->Set(env->context(), 0, entries).FromJust();
-  ret->Set(env->context(), 1, Boolean::New(env->isolate(), is_key_value))
-      .FromJust();
-  return args.GetReturnValue().Set(ret);
-}
 
-// Side effect-free stringification that will never throw exceptions.
-static void SafeToString(const FunctionCallbackInfo<Value>& args) {
-  auto context = args.GetIsolate()->GetCurrentContext();
-  args.GetReturnValue().Set(args[0]->ToDetailString(context).ToLocalChecked());
+  Local<Value> ret[] = {
+    entries,
+    Boolean::New(env->isolate(), is_key_value)
+  };
+  return args.GetReturnValue().Set(
+      Array::New(env->isolate(), ret, arraysize(ret)));
 }
 
 inline Local<Private> IndexToPrivateSymbol(Environment* env, uint32_t index) {
@@ -152,33 +161,9 @@ static void SetHiddenValue(const FunctionCallbackInfo<Value>& args) {
   args.GetReturnValue().Set(maybe_value.FromJust());
 }
 
-
-void StartSigintWatchdog(const FunctionCallbackInfo<Value>& args) {
-  int ret = SigintWatchdogHelper::GetInstance()->Start();
-  args.GetReturnValue().Set(ret == 0);
-}
-
-
-void StopSigintWatchdog(const FunctionCallbackInfo<Value>& args) {
-  bool had_pending_signals = SigintWatchdogHelper::GetInstance()->Stop();
-  args.GetReturnValue().Set(had_pending_signals);
-}
-
-
-void WatchdogHasPendingSigint(const FunctionCallbackInfo<Value>& args) {
-  bool ret = SigintWatchdogHelper::GetInstance()->HasPendingSignal();
-  args.GetReturnValue().Set(ret);
-}
-
-void SafeGetenv(const FunctionCallbackInfo<Value>& args) {
-  CHECK(args[0]->IsString());
-  Utf8Value strenvtag(args.GetIsolate(), args[0]);
-  std::string text;
-  if (!node::SafeGetenv(*strenvtag, &text)) return;
-  args.GetReturnValue()
-      .Set(String::NewFromUtf8(
-            args.GetIsolate(), text.c_str(),
-            NewStringType::kNormal).ToLocalChecked());
+void ArrayBufferViewHasBuffer(const FunctionCallbackInfo<Value>& args) {
+  CHECK(args[0]->IsArrayBufferView());
+  args.GetReturnValue().Set(args[0].As<ArrayBufferView>()->HasBuffer());
 }
 
 class WeakReference : public BaseObject {
@@ -209,30 +194,60 @@ class WeakReference : public BaseObject {
   SET_NO_MEMORY_INFO()
 
  private:
-  Persistent<Object> target_;
+  Global<Object> target_;
 };
 
+static void GuessHandleType(const FunctionCallbackInfo<Value>& args) {
+  Environment* env = Environment::GetCurrent(args);
+  int fd;
+  if (!args[0]->Int32Value(env->context()).To(&fd)) return;
+  CHECK_GE(fd, 0);
+
+  uv_handle_type t = uv_guess_handle(fd);
+  const char* type = nullptr;
+
+  switch (t) {
+    case UV_TCP:
+      type = "TCP";
+      break;
+    case UV_TTY:
+      type = "TTY";
+      break;
+    case UV_UDP:
+      type = "UDP";
+      break;
+    case UV_FILE:
+      type = "FILE";
+      break;
+    case UV_NAMED_PIPE:
+      type = "PIPE";
+      break;
+    case UV_UNKNOWN_HANDLE:
+      type = "UNKNOWN";
+      break;
+    default:
+      ABORT();
+  }
+
+  args.GetReturnValue().Set(OneByteString(env->isolate(), type));
+}
+
 void Initialize(Local<Object> target,
                 Local<Value> unused,
-                Local<Context> context) {
+                Local<Context> context,
+                void* priv) {
   Environment* env = Environment::GetCurrent(context);
 
 #define V(name, _)                                                            \
   target->Set(context,                                                        \
               FIXED_ONE_BYTE_STRING(env->isolate(), #name),                   \
-              Integer::NewFromUnsigned(env->isolate(), index++)).FromJust();
+              Integer::NewFromUnsigned(env->isolate(), index++)).Check();
   {
     uint32_t index = 0;
     PER_ISOLATE_PRIVATE_SYMBOL_PROPERTIES(V)
   }
 #undef V
 
-  target->DefineOwnProperty(
-    env->context(),
-    OneByteString(env->isolate(), "pushValToArrayMax"),
-    Integer::NewFromUnsigned(env->isolate(), NODE_PUSH_VAL_TO_ARRAY_MAX),
-    ReadOnly).FromJust();
-
 #define V(name)                                                               \
   target->Set(context,                                                        \
               FIXED_ONE_BYTE_STRING(env->isolate(), #name),                   \
@@ -247,18 +262,12 @@ void Initialize(Local<Object> target,
   env->SetMethod(target, "setHiddenValue", SetHiddenValue);
   env->SetMethodNoSideEffect(target, "getPromiseDetails", GetPromiseDetails);
   env->SetMethodNoSideEffect(target, "getProxyDetails", GetProxyDetails);
-  env->SetMethodNoSideEffect(target, "safeToString", SafeToString);
   env->SetMethodNoSideEffect(target, "previewEntries", PreviewEntries);
   env->SetMethodNoSideEffect(target, "getOwnNonIndexProperties",
                                      GetOwnNonIndexProperties);
+  env->SetMethodNoSideEffect(target, "getConstructorName", GetConstructorName);
 
-  env->SetMethod(target, "startSigintWatchdog", StartSigintWatchdog);
-  env->SetMethod(target, "stopSigintWatchdog", StopSigintWatchdog);
-  env->SetMethodNoSideEffect(target, "watchdogHasPendingSigint",
-                             WatchdogHasPendingSigint);
-
-  env->SetMethod(target, "safeGetenv", SafeGetenv);
-
+  env->SetMethod(target, "arrayBufferViewHasBuffer", ArrayBufferViewHasBuffer);
   Local<Object> constants = Object::New(env->isolate());
   NODE_DEFINE_CONSTANT(constants, ALL_PROPERTIES);
   NODE_DEFINE_CONSTANT(constants, ONLY_WRITABLE);
@@ -268,7 +277,15 @@ void Initialize(Local<Object> target,
   NODE_DEFINE_CONSTANT(constants, SKIP_SYMBOLS);
   target->Set(context,
               FIXED_ONE_BYTE_STRING(env->isolate(), "propertyFilter"),
-              constants).FromJust();
+              constants).Check();
+
+  Local<String> should_abort_on_uncaught_toggle =
+      FIXED_ONE_BYTE_STRING(env->isolate(), "shouldAbortOnUncaughtToggle");
+  CHECK(target
+            ->Set(env->context(),
+                  should_abort_on_uncaught_toggle,
+                  env->should_abort_on_uncaught_toggle().GetJSArray())
+            .FromJust());
 
   Local<String> weak_ref_string =
       FIXED_ONE_BYTE_STRING(env->isolate(), "WeakReference");
@@ -278,10 +295,12 @@ void Initialize(Local<Object> target,
   weak_ref->SetClassName(weak_ref_string);
   env->SetProtoMethod(weak_ref, "get", WeakReference::Get);
   target->Set(context, weak_ref_string,
-              weak_ref->GetFunction(context).ToLocalChecked()).FromJust();
+              weak_ref->GetFunction(context).ToLocalChecked()).Check();
+
+  env->SetMethod(target, "guessHandleType", GuessHandleType);
 }
 
 }  // namespace util
 }  // namespace node
 
-NODE_BUILTIN_MODULE_CONTEXT_AWARE(util, node::util::Initialize)
+NODE_MODULE_CONTEXT_AWARE_INTERNAL(util, node::util::Initialize)
